(*<*)
theory HoareMultiCall = IMP:
(*>*)

section{*Hoare-style program logic*}

text{*\label{sec:Hoare} This is the logic from the appendix, extended
by contexts. Again, the program logic is a partial correctness
logic.*}

subsection {*Assertions and their semantic validity*}

text{*Assertions state predicates, i.e.~are of type*}

types "HoareAssn" = "State \<Rightarrow> bool"

text{*Command $C$ satisfies specification $P,Q$ if all (terminating)
operational behaviours are covered by the specification.*}

constdefs Hoare_valid :: "HoareAssn \<Rightarrow> IMP \<Rightarrow> HoareAssn \<Rightarrow> bool"
                       (" \<Turnstile> { _ } _  { _ } " [100,100,100] 100)
"Hoare_valid P C Q \<equiv> (\<forall> s t . (s,C \<Down> t) \<longrightarrow> P s \<longrightarrow> Q t)"

text{*A variation of this property for the height-indexed operational
semantics,\ldots*}

constdefs Hoare_validn :: "nat \<Rightarrow> HoareAssn \<Rightarrow> IMP \<Rightarrow> HoareAssn \<Rightarrow> bool"
                        (" \<Turnstile>\<^sub>_ { _ } _ { _ }" [100,100,100,100] 100)
"Hoare_validn n P C Q \<equiv> 
   (\<forall> m . m \<le> n \<longrightarrow> (\<forall> s t . (s,C \<longrightarrow>\<^sub>m t) \<longrightarrow> P s \<longrightarrow> Q t))"

text{*\ldots plus the obvious relationships.*}
lemma Hoare_valid_validn: "\<Turnstile> {P} C {Q} \<Longrightarrow> \<Turnstile>\<^sub>n {P} C {Q}"
(*<*)
by (simp add: Hoare_valid_def Hoare_validn_def Sem_def, fastsimp)
(*>*)

lemma Hoare_validn_valid: "(\<forall> n . \<Turnstile>\<^sub>n {P}C{Q}) \<Longrightarrow> \<Turnstile> {P}C{Q}"
(*<*)
by (simp add: Hoare_valid_def Hoare_validn_def Sem_def, fastsimp)
(*>*)

lemma Hoare_lowerm: "\<lbrakk> \<Turnstile>\<^sub>n {P}C{Q}; m \<le> n \<rbrakk> \<Longrightarrow> \<Turnstile>\<^sub>m {P}C{Q}"
(*<*)
apply (simp add: Hoare_validn_def, clarsimp)
apply (rename_tac k s t)
apply (erule_tac x=k in allE, simp)
done
(*>*)

text{*Proof contexts are simply sets of assertions -- each entry
represents an assumption for the unnamed procedure. In particular, a
context is valid if each entry is satisfied by the method call
instruction.*}

types HoareCtxt = "(HoareAssn \<times> HoareAssn) set"

constdefs Ctxt_valid :: "HoareCtxt \<Rightarrow> bool"
                        (" \<Turnstile> _ " [100] 100)
"Ctxt_valid G \<equiv> \<forall> P Q . (P,Q) \<in> G \<longrightarrow> (\<Turnstile> {P}Call{Q})"

text{*Again, a relativised sibling \ldots*}

constdefs Ctxt_validn :: "nat \<Rightarrow> HoareCtxt \<Rightarrow> bool"
                         (" \<Turnstile>\<^sub>_ _  " [100,100] 100)
"Ctxt_validn n G \<equiv>
   \<forall> m . m \<le> n \<longrightarrow> (\<forall> P Q. (P,Q) \<in> G \<longrightarrow> ( \<Turnstile>\<^sub>m {P}Call{Q}) )"

text{*satisfies the obvious properties.*}

lemma Ctxt_valid_validn: "\<Turnstile> G \<Longrightarrow> \<Turnstile>\<^sub>n G"
(*<*)
apply (simp add: Ctxt_valid_def Ctxt_validn_def, clarsimp)
apply (rule Hoare_valid_validn) apply fast
done
(*>*)

lemma Ctxt_validn_valid: "(\<forall> n . \<Turnstile>\<^sub>n G) \<Longrightarrow> \<Turnstile> G"
(*<*)
apply (simp add: Ctxt_valid_def Ctxt_validn_def, clarsimp)
apply (rule Hoare_validn_valid) apply fast
done
(*>*)

lemma Ctxt_lowerm: "\<lbrakk> \<Turnstile>\<^sub>n G; m < n \<rbrakk> \<Longrightarrow> \<Turnstile>\<^sub>m G"
(*<*)
by (simp add: Ctxt_validn_def)
(*>*)

text{*A judgement is valid if the validity of the context implies that
of the commmand-assertion pair.*}

constdefs  valid :: "HoareCtxt \<Rightarrow> HoareAssn \<Rightarrow> IMP \<Rightarrow> HoareAssn \<Rightarrow> bool"
                    ("_ \<Turnstile> { _ } _  { _ }" [100,100,100,100] 100)
"valid G P C Q \<equiv>  (Ctxt_valid G \<longrightarrow> Hoare_valid P C Q)"

text{*And, again, a relatived notion of judgement validity.*}

constdefs  validn :: "HoareCtxt \<Rightarrow> nat \<Rightarrow> HoareAssn \<Rightarrow> IMP \<Rightarrow>
                      HoareAssn \<Rightarrow> bool"
                    ("_ \<Turnstile>\<^sub>_ { _ } _ { _ }" [100,100,100,100,100] 100)
"validn G n P C Q \<equiv>  (Ctxt_validn n G \<longrightarrow> Hoare_validn n P C Q)"

lemma validn_valid: "(\<forall> n . G \<Turnstile>\<^sub>n {P} C {Q}) \<Longrightarrow> G \<Turnstile> {P} C {Q}"
(*<*)
apply (simp add: valid_def validn_def, clarsimp)
apply (rule Hoare_validn_valid, clarsimp) 
apply (erule_tac x=n in allE, erule mp)
apply (erule Ctxt_valid_validn)
done
(*>*)

lemma ctxt_multi_consn: "\<lbrakk> \<Turnstile>\<^sub>n G;  \<Turnstile>\<^sub>n H \<rbrakk> \<Longrightarrow> \<Turnstile>\<^sub>n (H \<union> G)"
(*<*)
apply (simp add: Ctxt_validn_def)
done
(*>*)

lemma ctxt_consn: "\<lbrakk> \<Turnstile>\<^sub>n G;  \<Turnstile>\<^sub>n {P} Call {Q} \<rbrakk> \<Longrightarrow> \<Turnstile>\<^sub>n ({(P,Q)} \<union> G)"
(*<*)
apply (erule ctxt_multi_consn)
apply (simp add: Ctxt_validn_def) apply clarsimp 
apply (erule Hoare_lowerm) apply assumption
done
(*>*)


subsection{*Proof system*}

text{*The type of the judgements, plus some pretty-printing stuff.*}

consts Hoare_proof :: "(HoareCtxt \<times> HoareAssn \<times> IMP \<times> HoareAssn) set"
syntax Hoare_deriv :: "[HoareCtxt, HoareAssn, IMP, HoareAssn] \<Rightarrow> bool"
                     ("_ \<rhd> { _ } _ { _ }" [100,100,100,100] 100)
translations "G \<rhd> {P} C {Q}" == "(G,P,C,Q) \<in> Hoare_proof"

text{*Now, the proof rules. Note that the rule for CALL extends the
proof context by a \emph{set} of specifications. This generalizes the
usueal treatment but appears necessary for the interpretation of our
VDM logic: as each VDM judgement is interpreted as a set of Hoare
triples, a single application of the VDM-CALL rule adds multiple
Hoare-specifcations to the context. The rule of consequence is
slightly stronger than the one often presented in that it has a single
side condition which combines the weakening of the post-condition with
the strengthening of the pre-condition. Again this is beneficial for
embedding the corresponding VDM rule. Finally, we include the logical
rules {\sc HoareTrue} and {\sc HoareInter}, plus two variants of the
extract-exists rule -- only the one for states is actually used
later.*}

inductive Hoare_proof intros 
HoareSkip:   "G \<rhd> {P} Skip {P}"
HoareAssign: "G \<rhd> {(\<lambda> s . Q (update s x (evalE e s)))} (Assign x e) {Q}"
HoareComp:   "\<lbrakk> G \<rhd> {P} C1 {R}; G \<rhd> {R} C2 {Q}\<rbrakk> 
           \<Longrightarrow> G \<rhd> {P} (Comp C1 C2) {Q}"
HoareIff:    "\<lbrakk> G \<rhd> {P1} C1 {Q}; G \<rhd> {P2} C2 {Q} \<rbrakk>
            \<Longrightarrow> G \<rhd> {(\<lambda> s . (evalB b s \<longrightarrow> P1 s) \<and> 
                            ((\<not> evalB b s) \<longrightarrow> P2 s))} (Iff b C1 C2) {Q}"
(*HoareWhile:  "\<lbrakk>\<rhd> C : (\<lambda> s t . (evalB b s \<longrightarrow> I s t)); Trans I; Refl I\<rbrakk>
            \<Longrightarrow> \<rhd> (While b C) : (\<lambda> s t . I s t \<and> \<not> (evalB b t))"*)
HoareWhile:  "\<lbrakk> G \<rhd> {P'} C {P}; 
              \<forall> s . (\<not> evalB b s) \<longrightarrow> P s \<longrightarrow> Q s;
              \<forall> s . evalB b s \<longrightarrow> P s \<longrightarrow> P' s \<rbrakk>
            \<Longrightarrow> G \<rhd> {P} (While b C) {(\<lambda> s . Q s \<and> \<not> (evalB b s))}"
HoareMultiCall: 
   "\<lbrakk>\<forall> P' Q' . (P',Q') : GG \<longrightarrow> (GG \<union> G) \<rhd> {P'} body {Q'};
    (P,Q): GG\<rbrakk> \<Longrightarrow> G \<rhd> {P} Call {Q}"
HoareAx: "(P,Q) \<in> G \<Longrightarrow>  G \<rhd> {P} Call {Q}"
HoareTrue: "G \<rhd> {P} C {(\<lambda> s . True)}"
HoareConseqStrong:
   "\<lbrakk> G \<rhd> {P} C {Q}; \<forall> s t . P' s \<longrightarrow> (P s \<longrightarrow> Q t) \<longrightarrow> Q' t\<rbrakk>
     \<Longrightarrow> G \<rhd> {P'} C {Q'}"
extractexistsState:
"\<forall> (x::State) . G \<rhd> {(P x)} c {(Q x)}
  \<Longrightarrow> G \<rhd> { (\<lambda> s . \<exists> u . P u s)} c {(\<lambda> t . \<exists> u . Q u t)}"
extractexistsInt:
"\<forall> (x::int) . G \<rhd> {(P x)} c {(Q x)}
  \<Longrightarrow> G \<rhd> { (\<lambda> s . \<exists> u . P u s)} c {(\<lambda> t . \<exists> u . Q u t)}"
HoareInter: "\<lbrakk> G \<rhd> { P } C { Q1}; G \<rhd> { P } C { Q2 }\<rbrakk>
       \<Longrightarrow> G \<rhd> { P } C { (\<lambda> s . Q1 s \<and> Q2 s) }"

subsection{*Derivable rules*}

text{*First, the rule for single calls.*}

lemma HoareCall: "({(P,Q)} \<union> G) \<rhd> {P} body {Q} \<Longrightarrow> G \<rhd> {P} Call {Q}"
apply (rule_tac GG="{(P,Q)}" in HoareMultiCall)
  apply clarsimp 
  apply simp
done

text{*Next, the more standard rule of consequence.*}

lemma HoareConseq: 
  "\<lbrakk> G \<rhd> {P} C {Q}; \<forall> s . P' s \<longrightarrow> P s; \<forall> s .  Q s \<longrightarrow> Q' s\<rbrakk>
   \<Longrightarrow> G \<rhd> {P'} C {Q'}"
(*<*)
apply (erule HoareConseqStrong)
apply clarsimp 
done
(*>*)

(*In principle, one might even consider the following rule of
consequence: HoareConseqOpsem:

  "\<lbrakk> G \<rhd> {P} C {Q}; 
    \<forall> s t . P' s \<longrightarrow> s,C\<Down>t \<longrightarrow> (P s \<longrightarrow> s,C\<Down>t \<longrightarrow> Q t) \<longrightarrow> Q' t\<rbrakk>
    \<Longrightarrow> G \<rhd> {P'} C {Q'}"
 
Nobody seems to include the operational judgements in the side
condition, although this appears to me to do no harm, and occasionally
might be beneficial, since one can exploit the fact that the states
for which the side condition is required to hold are not arbitrary. I
suppose this rule can only be derived from HoareConseq using
completeness?  Rule HoareConseqOpsem should allow us to derive rule
VDMOpsem, which would in turn is useful for deriving the alternative
rule REL_False.  *)

text{*Again, we've also got a rule for extending contexts.*}

lemma HoareCtxtSub[rule_format]:
 "G \<rhd> {P} C {Q} \<Longrightarrow> \<forall> H .  G \<subseteq> H \<longrightarrow> H \<rhd> {P} C {Q}"
(*<*)
apply (erule Hoare_proof.induct)
apply clarsimp apply (rule HoareSkip)
apply clarsimp apply (rule HoareAssign)
apply clarsimp apply (rule_tac R=R in HoareComp) 
  apply (erule_tac x=H in allE, erule mp) apply assumption
  apply (erule_tac x=H in allE, erule mp) apply assumption
apply clarsimp apply (rule HoareIff)
  apply (erule_tac x=H in allE, erule mp) apply assumption
  apply (erule_tac x=H in allE, erule mp) apply assumption
apply clarsimp apply (rule_tac P'=P' in HoareWhile)
  apply (erule_tac x=H in allE, erule mp) apply assumption
  apply assumption
  apply assumption
(*apply clarsimp apply (rule HoareCall)
  apply (erule_tac x="{(P,Q)} \<union> H" in allE, erule mp) 
  apply fastsimp *)
apply clarsimp apply (rule_tac GG=GG in HoareMultiCall)
  apply clarsimp 
    apply (erule_tac x=P' in allE, erule_tac x=Q' in allE, clarsimp)
    apply (erule_tac x="GG \<union> H" in allE, erule mp) apply fastsimp
  apply assumption
apply clarsimp apply (rule HoareAx) apply fast
apply clarsimp apply (rule HoareTrue)
apply clarsimp apply (rule_tac P=P and Q=Q in HoareConseqStrong) 
  apply (erule_tac x=H in allE, erule mp) apply assumption 
  apply clarsimp
apply clarsimp 
  apply (rule extractexistsState) apply clarsimp
apply clarsimp 
  apply (rule extractexistsInt) apply clarsimp
apply clarsimp
  apply (erule_tac x=H in allE, erule impE, assumption)+ 
  apply (erule HoareInter) apply assumption
done
(*>*)

text{*The while-rule is in fact inter-derivable with the following rule.*}
lemma Hoare_While_derivable: 
   "G \<rhd> {(\<lambda> s . P s \<and> evalB b s)} C {P} \<Longrightarrow>
    G \<rhd> {P} While b C {(\<lambda> s . P s \<and> \<not> evalB b s)}"
(*<*)
apply (rule HoareConseq)
apply (erule_tac Q=P  in HoareWhile) apply simp apply simp apply simp apply simp 
done
(*>*)

text{*Here's the proof in the opposite direction.*}

lemma HoareWhile_derivable:
  "\<lbrakk> G \<rhd> {P'} C {P}; 
              \<forall> s . (\<not> evalB b s) \<longrightarrow> P s \<longrightarrow> Q s;
              \<forall> s . evalB b s \<longrightarrow> P s \<longrightarrow> P' s \<rbrakk>
            \<Longrightarrow> G \<rhd> {P} (While b C) {(\<lambda> s . Q s \<and> \<not> (evalB b s))}"
(*<*)
apply (rule_tac P=P in HoareConseq)
apply (rule Hoare_While_derivable)
apply (erule HoareConseq) apply clarsimp apply simp apply simp apply simp
done
(*>*)

text{*The following rule is an easy consequence of HoareConseqStrong,
and used in the embedding of the VDM logic.*}

lemma HoareCons: " G \<rhd> {(\<lambda> s . s=r)} C {(Q r)} \<Longrightarrow>
  G \<rhd> {P} C {(\<lambda> t . \<exists> s . P s \<and> (s = r \<longrightarrow> Q r t))}"
(*<*)
apply (erule HoareConseqStrong)
apply clarsimp 
apply (rule_tac x=s in exI,simp) 
done
(*>*)

text {*Finally, some simple variations of the rules for composition
and conditionals.*}

lemma HoareIff2:
"\<lbrakk> G \<rhd> {P} C1 {Q}; G \<rhd> {R} C2 {S}; 
   \<forall> s . T s \<longrightarrow> ((evalB b s \<longrightarrow> (P s \<and> (\<forall> t . Q t \<longrightarrow> U t)))) \<and> 
                  ((\<not> evalB b s) \<longrightarrow> (R s \<and> (\<forall> t . S t \<longrightarrow> U t)))\<rbrakk>
\<Longrightarrow> G \<rhd> {T} (Iff b C1 C2) {U}"
(*<*)
apply (rule HoareConseqStrong)
apply (rule_tac Q=U in HoareIff) 
  apply (erule HoareConseqStrong) apply (subgoal_tac "\<forall> s t. (T s \<and> evalB b s) \<longrightarrow> (P s \<longrightarrow> Q t) \<longrightarrow> U t") apply assumption
  apply clarsimp 
apply (erule HoareConseqStrong) apply (subgoal_tac "\<forall> s t. (T s \<and> \<not> evalB b s) \<longrightarrow> (R s \<longrightarrow> S t) \<longrightarrow> U t") apply assumption
  apply clarsimp
apply clarsimp 
done
(*>*)

lemma HoareCompDecoupled: 
  "\<lbrakk> G \<rhd> {P} C1 {R}; G \<rhd> {S} C2 {Q}; \<forall> s . R s \<longrightarrow> S s\<rbrakk> 
           \<Longrightarrow> G \<rhd> {P} (Comp C1 C2) {Q}"
(*<*)
apply (rule HoareComp) apply assumption 
apply (erule HoareConseq) apply assumption apply simp
done
(*>*)

(*The following rule is derivable of we use HoareConseqOpsem, but not
needed in what follows.

lemma 
   "\<lbrakk> G \<rhd> {P} C1 {(\<lambda> t . \<exists> s . P s \<and> s,C1\<Down>t \<and> R t)}; G \<rhd> {S} C2 {Q};
      (\<forall> r . (\<exists> u . u,C1\<Down>r \<and> P u \<and> R r) \<longrightarrow> S r)\<rbrakk> 
   \<Longrightarrow> G \<rhd> {P} (Comp C1 C2) {Q}"
apply (rule_tac Q=Q in HoareConseq)
apply (rule HoareComp) apply assumption 
apply (erule HoareConseq2) apply clarsimp apply (erule_tac x=s in allE)  apply (erule impE)  apply (rule_tac x=sa in exI, simp) apply fast 
apply simp+
done
*)

subsection{*Soundness*}
(*<*)
lemma MAX:"Suc (max k m) \<le> n \<Longrightarrow> k \<le> n \<and> m \<le> n"
by (simp add: max_def, case_tac "k \<le> m", simp+)
(*>*)


text{*An auxiliary lemma stating the soundness of the while rule. Its
proof is by induction on $n$.*}

lemma SoundWhile[rule_format]:
" (\<forall> m . G \<Turnstile>\<^sub>m   { P' } C { P }) \<longrightarrow> 
  (\<forall>s. \<not> evalB b s \<longrightarrow> P s \<longrightarrow> Q s) \<longrightarrow> 
  (\<forall>s. evalB b s \<longrightarrow> P s \<longrightarrow> P' s)
  \<longrightarrow> G \<Turnstile>\<^sub>n { P } While b C { (\<lambda>u. Q u \<and> \<not> evalB b u) }"
(*<*)
apply (induct n)
apply (simp add: validn_def Hoare_validn_def, clarsimp) 
  apply (drule Sem_no_zero_height_derivs) apply simp 
apply clarsimp apply (simp add: validn_def Hoare_validn_def, clarsimp)  
  apply (erule Sem_eval_cases)
  prefer 2 apply clarsimp 
  apply clarsimp 
   apply (erule_tac x=n in allE, erule impE) apply (erule Ctxt_lowerm) apply simp
   apply (rotate_tac -1)
   apply (erule_tac x=ma in allE, clarsimp) 
   apply(erule impE) apply (erule Ctxt_lowerm) apply simp 
   apply (erule_tac x=na in allE, clarsimp)
done
(*>*)

text{*Similarly, an auxiliary lemma for procedure invocations. As a
warm-up, first the version for a single specification. *}

lemma SoundCall[rule_format]:
  "\<lbrakk>\<forall>m. ({(P, Q)} \<union> G) \<Turnstile>\<^sub>m { P } body { Q }\<rbrakk> \<Longrightarrow>  \<Turnstile>\<^sub>n G \<longrightarrow> \<Turnstile>\<^sub>n {P} Call {Q}"
(*<*)
apply (induct_tac n)
apply (simp add: Hoare_validn_def, clarsimp) 
  apply (drule Sem_no_zero_height_derivs) apply simp 
apply clarsimp
  apply (drule Ctxt_lowerm) apply (subgoal_tac "n < Suc n", assumption) apply simp apply clarsimp
  apply (drule ctxt_consn) apply assumption
  apply (simp add: Hoare_validn_def, clarsimp)
  apply (erule Sem_eval_cases, clarsimp) apply (erule_tac x=n in allE) 
     apply (simp add: validn_def) apply (erule thin_rl) apply (erule thin_rl) 
      apply (simp add: Hoare_validn_def)
done
(*>*)

text{*The soundness proof actually employs the more general rulefor
multiple specifications.*}

lemma SoundMultiCall[rule_format]:
  "(\<forall>m. \<forall> P' Q' . (P',Q'):GG \<longrightarrow> (GG \<union> G) \<Turnstile>\<^sub>m { P' } body { Q' }) \<longrightarrow>
   (\<forall> P Q .  (P,Q):GG \<longrightarrow>  \<Turnstile>\<^sub>n G \<longrightarrow> \<Turnstile>\<^sub>n {P} Call {Q})"
(*<*)
apply (induct_tac n)
apply (simp add: Hoare_validn_def, clarsimp) 
  apply (drule Sem_no_zero_height_derivs) apply simp 
apply clarsimp
  apply (drule Ctxt_lowerm) apply (subgoal_tac "n < Suc n", assumption) apply simp apply clarsimp
  apply (drule_tac H=GG in ctxt_multi_consn) apply (simp add: Ctxt_validn_def)
     apply clarsimp 
     apply (erule_tac x=Pa in allE, erule_tac x=Qa in allE,clarsimp)
     apply (erule Hoare_lowerm) apply assumption
  apply (simp add: Hoare_validn_def, clarsimp)
  apply (erule Sem_eval_cases, clarsimp) apply (erule_tac x=n in allE) 
     apply (simp add: validn_def)
        apply (rotate_tac 1) apply (erule thin_rl) apply (erule thin_rl)
     apply (erule_tac x=P in allE, erule_tac x=Q in allE,clarsimp) 
      apply (simp add: Hoare_validn_def)
done
(*>*)

text{*The heart of the soundness proof is the following lemma which is
proven by induction on the judgement $G \rhd \{P\} C \{Q\}$.*}

lemma Hoare_Sound_n: "G \<rhd> {P} C {Q} \<Longrightarrow> (\<forall> n . G \<Turnstile>\<^sub>n {P} C {Q})"
(*<*)
apply (erule Hoare_proof.induct)
apply (simp add: validn_def Hoare_validn_def)
  apply(clarsimp, erule Sem_eval_cases, simp)
apply (simp add: validn_def Hoare_validn_def)
  apply(clarsimp, erule Sem_eval_cases, simp)
apply (simp add: validn_def Hoare_validn_def)
  apply(clarsimp, erule Sem_eval_cases, clarsimp)
  apply (drule MAX, clarsimp) 
apply (simp add: validn_def Hoare_validn_def)
  apply(clarsimp, erule Sem_eval_cases, clarsimp)
   apply (erule thin_rl, rotate_tac 1, erule thin_rl, erule thin_rl)
     apply (erule_tac x=n in allE, clarsimp, erule_tac x=na in allE, clarsimp)
   apply (erule thin_rl, erule thin_rl)
     apply (erule_tac x=n in allE, clarsimp, erule_tac x=na in allE, clarsimp)
apply clarsimp
  apply (rule SoundWhile) apply fast apply simp apply simp 
apply (simp (no_asm) add: validn_def, clarsimp)
  apply (rule_tac GG=GG and G=G in SoundMultiCall) apply simp 
   apply assumption apply assumption
(*
apply (simp (no_asm) add: validn_def, clarsimp)
  apply (rule SoundCall) prefer 2 apply assumption apply simp *)
apply (simp add: Ctxt_validn_def validn_def) 
(*HoareTrue*)
apply (simp add: validn_def Hoare_validn_def) 
(*ConseqStrong*)
apply (simp add: validn_def Hoare_validn_def Sem_def, clarsimp) 
(*  apply (erule_tac x=s in allE, clarsimp)
  apply (erule_tac x=t in allE, erule impE) apply (rule, assumption)
  apply simp*)
apply (simp add: validn_def Hoare_validn_def Sem_def, clarsimp)
  apply (rule_tac x=u in exI, simp)  
apply (simp add: validn_def Hoare_validn_def Sem_def, clarsimp)
  apply (rule_tac x=u in exI, simp)   
(*HoareInter*)
apply (simp add: validn_def Hoare_validn_def)
done
(*>*)

text{*Thus, we obtain soundness in contexts*}

theorem Hoare_Sound: "G \<rhd> {P} C {Q} \<Longrightarrow> G \<Turnstile> {P} C {Q}"
(*<*)
by (drule Hoare_Sound_n, erule validn_valid) 
(*>*)

text{*a simple corollary of which is soundness w.r.t.~an empty context.*}

lemma Hoare_Sound_emptyCtxt:"{} \<rhd> {P} C {Q} \<Longrightarrow> \<Turnstile> {P} C {Q}"
(*<*)
apply (drule Hoare_Sound)
apply (simp add: valid_def, erule mp) 
apply (simp add: Ctxt_valid_def)
done
(*>*)

(*<*)
end
(*>*)
(*
subsection{-Context elimination-}

(-<-)
constdefs CutAuxProp::"HoareCtxt \<Rightarrow> HoareAssn \<Rightarrow> IMP \<Rightarrow> HoareAssn \<Rightarrow> bool"
"CutAuxProp DD P C Q \<equiv> 
  \<forall> G PP QQ D . (DD = (insert (PP,QQ) D) \<longrightarrow> G \<rhd> {PP} Call {QQ} \<longrightarrow> (G \<subseteq> D) \<longrightarrow> D \<rhd> {P} C {Q})"

lemma CutAux1:"(DD \<rhd> {P} C {Q}) \<Longrightarrow> CutAuxProp DD P C Q"
apply (erule Hoare_proof.induct)
apply (simp_all add: add: CutAuxProp_def)
apply clarify
apply (fast intro: HoareSkip)
apply (fast intro: HoareAssign)
apply clarsimp 
  apply (erule_tac x=Ga in allE) apply (erule_tac x=Ga in allE)
  apply (erule_tac x=PP in allE) apply (erule_tac x=PP in allE)
  apply (erule_tac x=QQ in allE) apply (erule_tac x=QQ in allE)
  apply (erule_tac x=D in allE, simp) apply (erule_tac x=D in allE, simp)
  apply (rule HoareComp) apply assumption+
apply clarsimp 
  apply (erule_tac x=Ga in allE) apply (erule_tac x=Ga in allE)
  apply (erule_tac x=PP in allE) apply (erule_tac x=PP in allE)
  apply (erule_tac x=QQ in allE) apply (erule_tac x=QQ in allE)
  apply (erule_tac x=D in allE, simp) apply (erule_tac x=D in allE, simp)
  apply (rule HoareIff) apply assumption+
apply clarsimp 
  apply (erule_tac x=Ga in allE) 
  apply (erule_tac x=PP in allE) 
  apply (erule_tac x=QQ in allE)
  apply (erule_tac x=D in allE, simp) 
  apply (rule HoareWhile) apply assumption+
(-apply clarsimp 
  apply (erule_tac x=Ga in allE) 
  apply (erule_tac x=PP in allE) 
  apply (erule_tac x=QQ in allE)
  apply (erule_tac x="insert (P,Q) D" in allE, simp) 
  apply (rule HoareCall) apply fastsimp-)
apply clarsimp 
  apply (erule_tac x=P in allE, erule_tac x=Q in allE,clarsimp)
  apply (erule_tac x=Ga in allE) 
  apply (erule_tac x=PP in allE) 
  apply (erule_tac x=QQ in allE)
  apply (erule_tac x="(GG \<union> D)" in allE, simp) 
  apply (erule impE) apply fast
  apply (rule HoareMultiCall) apply fastsimp
apply clarsimp
  apply (erule disjE)
  apply clarsimp apply (erule HoareCtxtSub) apply assumption
  apply (erule HoareAx)
(-
apply clarsimp
  apply (rule_tac P=P and Q=Q in HoareConseq) 
  apply (erule_tac x=Ga in allE) apply (erule_tac x=PP in allE)
   apply (erule_tac x=QQ in allE) apply (erule_tac x=D in allE, clarsimp)
  apply assumption
  apply assumption
-)
apply clarsimp
  apply (rule_tac P=P and Q=Q in HoareConseq1) 
  apply (erule_tac x=Ga in allE) apply (erule_tac x=PP in allE)
   apply (erule_tac x=QQ in allE) apply (erule_tac x=D in allE, clarsimp)
  apply fast
apply clarsimp
  apply (rule extractexistsState, rule)
  apply (erule_tac x=x in allE, erule conjE)
  apply (erule_tac x=Ga in allE) apply (erule_tac x=PP in allE)
   apply (erule_tac x=QQ in allE) apply (erule_tac x=D in allE, clarsimp)
apply clarsimp
  apply (rule extractexistsInt, rule)
  apply (erule_tac x=x in allE, erule conjE)
  apply (erule_tac x=Ga in allE) apply (erule_tac x=PP in allE)
   apply (erule_tac x=QQ in allE) apply (erule_tac x=D in allE, clarsimp)
done
(->-)

(-lemma CutAux: 
  "(DD \<rhd> C : A) \<Longrightarrow>
  (\<forall> G P D . (DD = (insert P D) \<longrightarrow> G \<rhd> Call :P \<longrightarrow> (G \<subseteq> D) \<longrightarrow> D \<rhd> C:A))"
by (drule CutAux1, simp add: CutAuxProp_def)
-)
lemma CutAux: 
  "\<lbrakk>DD \<rhd> {P} C {Q}; DD = insert (PP,QQ) D; 
    G \<rhd> {PP} Call {QQ}; G \<subseteq> D\<rbrakk> \<Longrightarrow> D \<rhd> {P} C {Q}"
(-<-)
by (drule CutAux1, simp add: CutAuxProp_def)
(->-)

lemma cut:"\<lbrakk> G \<rhd> {PP} Call {QQ} ; (insert (PP,QQ) G) \<rhd> {P} C {Q} \<rbrakk> \<Longrightarrow> G \<rhd> {P} C {Q}"
(-<-)
apply (rotate_tac -1, drule CutAux)
apply (fastsimp, assumption)
apply (simp, assumption)
done
(->-)

text{-We call context $G$ verified if all entries are justified by
derivations for the procedure body.-}

constdefs verified::"HoareCtxt \<Rightarrow> bool"
"verified G \<equiv> (\<forall> P Q . (P,Q):G \<longrightarrow> G \<rhd> {P} body {Q})"

text{-The property is preserved by sub-contexts-}

lemma verified_preserved: "\<lbrakk>verified G; (P,Q):G\<rbrakk> \<Longrightarrow> verified (G - {(P,Q)})"
(-<-)
apply (simp add: verified_def, (rule allI)+, rule)
apply (subgoal_tac "(G - {(P,Q)}) \<rhd> {P} Call {Q}")
(-now use the subgoal (G - {(P,Q)}) \<rhd> {P} Call {Q} to prove the goal-)
apply (subgoal_tac "G = insert (P,Q) (G - {(P,Q)})") prefer 2 apply fast
apply (erule_tac x=Pa in allE) apply (erule_tac x=Qa in allE)
  apply (erule impE, simp)
  apply (erule cut)  apply simp
(- now prove the subgoal (G - {(P,Q)}) \<rhd>  {P} Call {Q} -)
  apply (erule_tac x=P in allE, erule_tac x=Q in allE, clarsimp)
  apply (rule HoareCall) apply simp apply (erule HoareCtxtSub) apply fast
done
(->-)

(-<-)
lemma SetNonSingleton:
  "\<lbrakk>G \<noteq> {A}; A \<in> G\<rbrakk> \<Longrightarrow> \<exists>B. B \<noteq> A \<and> B \<in> G"
by auto

lemma MUTREClemma[rule_format]: 
"\<forall> G . ((finite G \<and> card G = n \<and> verified G \<and> (P,Q): G) \<longrightarrow> {} \<rhd> {P} Call {Q})"
apply (induct n)
(-case n=0-)
apply clarsimp
(-Case n>0-)
apply clarsimp
apply (case_tac "G = {(P,Q)}")
apply clarsimp
apply (erule_tac x="{(P,Q)}" in allE)
apply (clarsimp, simp add:verified_def)
apply (rule HoareCall, clarsimp)
(-Case G has more entries than (P,Q)-)
apply (drule SetNonSingleton, assumption) 
(- use the fact that there is another entry (PP,QQ):G-)
apply clarsimp apply (rename_tac n G PP QQ)
apply (subgoal_tac "(G - {(PP,QQ)}) \<rhd> {PP} Call {QQ}")
apply (erule_tac x="G - {(PP,QQ)}" in allE)
apply (rotate_tac -1, erule impE) apply (simp add: verified_preserved) 
apply (rotate_tac -1, erule impE) apply (simp add: card_Diff_singleton) 
apply simp
(-the proof for (G - {(PP,QQ)}) \<rhd> {PP} Call {QQ} -)
apply (simp add: verified_def)
apply (rotate_tac 3) apply (erule_tac x=PP in allE, erule_tac x=QQ in allE, simp)
apply (rule HoareCall) apply simp apply (erule HoareCtxtSub) apply fast
done
(->-)

text{-The \emph{Mutrec} (we use this name despite the existence of
only a single procedure) rule allows to eliminate verified (finite)
contexts.-}

theorem MUTREC:
"\<lbrakk> finite G; card G = n; verified G; (P,Q) : G \<rbrakk> \<Longrightarrow> {} \<rhd> {P} Call {Q}"
(-<-) 
by (rule MUTREClemma, fast)
(->-)

(-<-)
end
(->-)
*)