(*<*)
theory IMP = Main:
(*>*)
section {*The language IMP *}

text{*The language defined here is a superset of the language
introduced in the RHL paper and includes the simple procedure call
instruction. In fact, the file is -- apart from cosmetic changes --
identical to the one from the formalization of our CSF2007 paper.*}

subsection{*Syntax*} 

text{*We start from unspecified categories of program variables and
values. *}

datatype Var = var int
datatype Val = iVal int

text{*Arithmetic expressions are inductively built up from variables,
values, and binary operators which are modeled as meta-logical
functions over values. Similarly, boolean expressions are built up
from arithmetic expressions using binary boolean operators which are again
modeled as functions of the ambient logic HOL.*}

datatype Expr =
   varE Var
 | valE Val
 | opE "Val \<Rightarrow> Val \<Rightarrow> Val" Expr Expr

datatype BExpr = compB "Val \<Rightarrow> Val \<Rightarrow> bool" Expr Expr

text{*Commands are the usual ones for an imperative language (see for
example Winskel's book), plus the command $\mathit{Call}$ which stands
for the invocation of a single (unnamed, parameterless) procedure.*}

datatype IMP =
    Skip 
  | Assign Var Expr
  | Comp IMP IMP
  | While BExpr IMP
  | Iff BExpr IMP IMP
  | Call

text{*The body of this procedure is identified by the following
constant.*}

consts body :: IMP

subsection{*Dynamic semantics*}

text{*For the language {\bf IMP}, states are simply given by stores -
in our case, (total) HOL functions mapping program variables to
values.*}

types State = "Var \<Rightarrow> Val"

constdefs update :: "State \<Rightarrow> Var \<Rightarrow> Val \<Rightarrow> State"
"update s x v \<equiv> (\<lambda> y . if x=y then v else s y)"

text{*The evaluation of expressions is defined inductively, as
standard.*}

consts evalE::"Expr \<Rightarrow> State \<Rightarrow> Val"
primrec
"evalE (varE x) s = s x" 
"evalE (valE v) s = v"
"evalE (opE f e1 e2) s = f (evalE e1 s) (evalE e2 s)"

consts evalB::"BExpr \<Rightarrow> State \<Rightarrow> bool"
primrec
"evalB (compB f e1 e2) s = f (evalE e1 s) (evalE e2 s)"

subsection{*Operational semantics*}

text{*The operational semantics is a standard big-step relation, with
a height index that facilitates the Hofmann-Kleymann-Nipkow-style
soundness proof of the program logic.*}

consts Semn :: "(State \<times> IMP \<times> nat \<times> State) set"

syntax Semn_  :: "[State, IMP, nat, State] \<Rightarrow> bool"   (" _ , _ \<longrightarrow>\<^sub>_  _ ")

translations "s,c \<longrightarrow>\<^sub>n t" == "(s,c,n,t) : Semn"

inductive Semn intros
 SemSkip:  "s,Skip \<longrightarrow>\<^sub>1 s"
 SemAssign:"\<lbrakk> t = update s x (evalE e s)\<rbrakk> \<Longrightarrow> s,(Assign x e) \<longrightarrow>\<^sub>1 t"
 SemComp:  "\<lbrakk> s,c1 \<longrightarrow>\<^sub>n r; r,c2 \<longrightarrow>\<^sub>m t; k=(max n m)+1\<rbrakk>
          \<Longrightarrow> s,(Comp c1 c2) \<longrightarrow>\<^sub>k t"
 SemWhileT:"\<lbrakk>evalB b s; s,c \<longrightarrow>\<^sub>n r; r,(While b c) \<longrightarrow>\<^sub>m t; 
             k=((max n m)+1)\<rbrakk>
          \<Longrightarrow> s,(While b c) \<longrightarrow>\<^sub>k t"
 SemWhileF:"\<lbrakk>\<not> (evalB b s); t=s\<rbrakk> \<Longrightarrow> s,(While b c) \<longrightarrow>\<^sub>1 t"
 SemTrue:  "\<lbrakk>evalB b s; s,c1 \<longrightarrow>\<^sub>n t\<rbrakk> \<Longrightarrow> s,(Iff b c1 c2) \<longrightarrow>\<^sub>(n+1) t"
 SemFalse: "\<lbrakk>\<not> (evalB b s); s,c2 \<longrightarrow>\<^sub>n t\<rbrakk> 
          \<Longrightarrow> s,(Iff b c1 c2) \<longrightarrow>\<^sub>(n+1) t"
 SemCall:  "\<lbrakk> s,body \<longrightarrow>\<^sub>n t\<rbrakk> \<Longrightarrow> s,Call \<longrightarrow>\<^sub>(n+1) t"

text{*Often, the height index does not matter\footnote{Using
B.~Gregoire's technique we could probably eliminate the index
completely\ldots}.*}

constdefs
 Sem :: "[State, IMP, State] \<Rightarrow> bool"  ("_ , _ \<Down> _ " 1000)
 "s,c \<Down> t \<equiv> (\<exists> n. (s,c \<longrightarrow>\<^sub>n t))"

text{*Inductive elimination rules for the (indexed) dynamic semantics:*}

inductive_cases Sem_eval_cases: 
 "s,Skip \<longrightarrow>\<^sub>n t"
 "s,(Assign x e) \<longrightarrow>\<^sub>n t"
 "s,(Comp c1 c2) \<longrightarrow>\<^sub>n t"
 "s,(While b c) \<longrightarrow>\<^sub>n t"
 "s,(Iff b c1 c2) \<longrightarrow>\<^sub>n t"
 "s, Call \<longrightarrow>\<^sub>n t"

(*<*)
lemma Sem_no_zero_height_derivsAux: "\<forall> s t. ((s, c \<longrightarrow>\<^sub>0 t) \<longrightarrow> False)"
by (induct_tac c, auto elim: Sem_eval_cases)
(*>*)

text{*An induction on $c$ shows that no derivations of height
$0$ exist.*}

lemma Sem_no_zero_height_derivs: "(s, c \<longrightarrow>\<^sub>0 t) \<Longrightarrow> False"
(*<*)by (insert Sem_no_zero_height_derivsAux, fastsimp)(*>*)

(*<*)
lemma SemnDeterm[rule_format]:
"(s, c \<longrightarrow>\<^sub>n t) \<Longrightarrow> (\<forall> r m . (s, c \<longrightarrow>\<^sub>m r) \<longrightarrow> m=n \<and> r=t)"
apply (erule Semn.induct)
apply(clarsimp, elim Sem_eval_cases, simp)
(*Assign*)
apply (rule allI)+ apply rule
apply(elim Sem_eval_cases)
apply simp
(*Comp*)
apply (rule allI)+ apply rule
apply(elim Sem_eval_cases)
apply simp
(*WhileT*)
apply (rule allI)+ apply rule
apply (rotate_tac 3) apply (erule thin_rl)
apply (erule Semn.elims) apply clarify+
  apply (rotate_tac -4) 
  apply (erule_tac x=rb in allE) apply (erule_tac x=na in allE) apply clarsimp 
  apply clarify+
(*WhileF*)
apply (erule Semn.elims) apply clarify+
  apply simp
  apply clarify+
(*True*)
apply(elim Sem_eval_cases) 
apply simp
apply fast
(*False*)
apply clarify
apply(elim Sem_eval_cases) 
apply fast
apply simp
(*Call*)
apply clarify
apply(elim Sem_eval_cases) 
apply simp
done
(*>*)

text{*The proof of determinism is by induction on the 
      (indexed) operational semantics.*}
lemma SemDeterm: "\<lbrakk>s, c \<Down> t; s, c \<Down> r\<rbrakk> \<Longrightarrow> r=t"
(*<*)
apply (simp add: Sem_def, clarsimp)
apply (drule SemnDeterm, assumption)
apply simp
done
(*>*)

(*<*)
end
(*>*)